from typing import Callable, List, Union
import random

import cv2
import numpy as np

from vcap.caching import cache


class DetectionPrediction:
    def __init__(self, class_name: str,
                 class_id: int,
                 rect: List[Union[int, float]],
                 confidence: float):
        """An object detection label. Basically a bounding box classifying
        some portion of an image.

        :param class_name: The class name that was predicted
                           (Price_Tag, Shelf_Empty, Person)
        :param class_id: The class number used by the network
        :param rect: [xmin, ymin, xmax, ymax] format
        :param confidence: The confidence in percent from 0-1, representing
            how confident the network is with the prediction
        """
        self._color = None

        # The full rect, (x1, y1, x2, y2) format
        self.rect = tuple(rect)

        self.name = str(class_name)
        self.confidence = float(confidence)
        self.class_num = int(class_id)

        assert 0. <= self.confidence <= 1, "Confidence must be between 0 and 1!"

    def __str__(self):
        return (str(self.name) + " " +
                str(self.rect) + " " +
                str(round(self.confidence, 2)))

    def __repr__(self):
        return self.__str__()

    @property
    def p1(self):
        """ Top Left of the rect"""
        return tuple(self.rect[:2])

    @property
    def p2(self):
        """ Top Right of the rect"""
        return tuple(self.rect[2:])

    @property
    def color(self):
        if self._color is None:
            # Generate a color based on the class name
            rand_seed = random.Random(self.name)
            self._color = tuple([rand_seed.randint(0, 255) for i in range(3)])

        return self._color


class ClassificationPrediction:
    def __init__(self,
                 class_scores: np.ndarray,  # floats
                 class_names: List[str]):
        """A classification of an entire image.

        :param class_scores: A list of scores for each class
        :param class_names: A list of class names, same length as class scores.
        Each index corresponds to an index in class_scores
        """
        self.class_scores = class_scores
        self.class_names = class_names

    def __iter__(self):
        """Iterate over class_name and class_score"""
        for name, score in zip(self.class_names, self.class_scores):
            # Ensure that no numpy floats get returned (causing bad errors)
            yield str(name), float(score)

    def __str__(self):
        return f"{self.name} {self.confidence:.2f}"

    def __repr__(self):
        return self.__str__()

    @property
    @cache
    def name(self) -> str:
        """The class name that was predicted with the highest confidence"""
        return self.class_names[self.class_num]

    @property
    @cache
    def class_num(self) -> int:
        """The class number used by the network for the class with the highest
        confidence"""
        return int(np.argmax(self.class_scores))

    @property
    @cache
    def confidence(self) -> float:
        """The confidence in percent from 0-1, representing how confident the
        network is with its highest confidence prediction"""
        return float(self.class_scores[self.class_num])


class EncodingPrediction:
    def __init__(self, encoding_vector: np.ndarray,
                 distance_func: Callable):
        """

        :param encoding_vector: The encoding prediction
        :param distance_func: A function with args (encoding, other_encodings)
        returns the distance from one encoding to all the other_encodings in
        """
        self.vector = encoding_vector
        self.distance = distance_func


class DensityPrediction:
    def __init__(self, density_map):
        self.map = density_map
        self.count = np.sum(density_map)

    def __str__(self):
        return str(self.count)

    def __repr__(self):
        return self.__str__()

    def resized_map(self, new_size):
        """
        Returns a resized density map, where the sum of each value is still the
        same count
        :param new_size: (new width, new height)"""

        new_map = cv2.resize(self.map.copy(), new_size)
        cur_count = np.sum(new_map)

        # Avoid dividing by zero
        if cur_count == 0:
            return new_map

        scale = self.count / cur_count
        new_map *= scale
        return new_map


class SegmentationPrediction:
    def __init__(self, segmentation, label_map):
        """
        :param segmentation: 2d segmented image where the value is the label num
        :param label_map: Formatted as {"1" : {"label" : "chair", "color" : [12, 53, 100]}}
        """
        self.segmentation = segmentation
        self.label_map = label_map

    def colored(self):
        """Convert segmented image to RGB  image using label_map"""

        color_img = np.zeros(
            (self.segmentation.shape[0], self.segmentation.shape[1], 3),
            dtype=np.uint8)

        # Replace segment values with color pixels using label_map values
        for label_num, label in self.label_map.items():
            color_img[self.segmentation == int(label_num)] = np.array(
                label["color"], dtype=np.uint8)

        return color_img


class DepthPrediction:
    def __init__(self, depth_prediction):
        """
        :param depth_prediction: 2d image where the value is the depth
        """
        self.depth_prediction = depth_prediction

    def normalized(self):
        """Convert segmented image to RGB  image using label_map"""

        # Scale results to max at 255 for image display
        max_distance = np.max(self.depth_prediction)
        pred = 255 * self.depth_prediction // max_distance

        # Convert results to uint8
        pred = pred.astype(np.uint8, copy=True)

        # Do fancy coloring
        pred = cv2.applyColorMap(pred, cv2.COLORMAP_JET)

        return pred



import json

import tensorflow as tf


def parse_tf_model_bytes(model_bytes,
                         device: str = None,
                         session_config: tf.ConfigProto = None):
    """

    :param model_bytes: The bytes of the model to load
    :param device_id: The device that this model should be loaded onto
    :param session_config: Configuration options for multiple monitors
    :return:
    """

    # Load the model as a graph
    detection_graph = tf.Graph()
    with detection_graph.as_default():
        # Load a (frozen) Tensorflow model from memory
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(model_bytes)

        with tf.device(device):
            tf.import_graph_def(graph_def,
                                input_map=None,
                                return_elements=None,
                                producer_op_list=None,
                                name='')

    if session_config is None:
        session_config = tf.ConfigProto()

    if device is not None:
        # allow_soft_placement lets us remap GPU only ops to GPU, and doesn't
        # crash for non-gpu only ops (it will place those on CPU, instead)
        session_config.allow_soft_placement = True

    # Create a session for later use
    persistent_sess = tf.Session(graph=detection_graph,
                                 config=session_config)

    return detection_graph, persistent_sess


def parse_dataset_metadata_bytes(dataset_metadata_bytes):
    """ Loads the label_map from the dataset_metadata bytes """
    dataset_metadata = json.loads(dataset_metadata_bytes.decode("utf-8"))

    # Convert the keys of the label map to integer values, instead of strings
    label_map = dataset_metadata["label_map"]
    label_map = {int(k): v for k, v in label_map.items()}
    return label_map



import cv2
import numpy as np
import tensorflow as tf

from .load_utils import parse_tf_model_bytes
from .predictions import DensityPrediction
from .base_tensorflow import BaseTFBackend


class CrowdDensityCounter(BaseTFBackend):
    """Runs object detection using a given model, meaning that it takes an
    image as input and returns bounding boxes with classifications as output.
    """

    def __init__(self, model_bytes,
                 device: str=None,
                 session_config: tf.ConfigProto=None):
        """
        :param model_bytes: Model file data, likely a loaded *.pb file
        :param device: The device to run the model on
        :param session_config: Model configuration options
        """
        super().__init__()
        self.graph, self.session = parse_tf_model_bytes(model_bytes,
                                                        device,
                                                        session_config)

        self.input = self.graph.get_tensor_by_name("input:0")
        self.output = self.graph.get_tensor_by_name("Threshold_32/Relu:0")

    def batch_predict(self, img_bgr) -> DensityPrediction:
        """Takes a numpy BGR image of the format that OpenCV gives, and returns
        predicted labels in the form of a list of Detection objects
        from predictions.py

        :img_bgr: An OpenCV image in BGR format (the default)
        :return: List of labels
        """
        preprocessed = preprocess(img_bgr)

        # Run the network
        feed = {self.input: preprocessed}
        density_map = self.session.run([self.output], feed)
        density_map = np.squeeze(density_map)

        return DensityPrediction(density_map)


def preprocess(img_bgr):
    # Convert the image to grayscale if it wasn't already
    if len(img_bgr.shape) == 3:
        gray = cv2.cvtColor(img_bgr, cv2.COLOR_BGR2GRAY)
        gray = gray.astype(np.float32, copy=False)
    elif len(img_bgr.shape) == 2:
        gray = img_bgr
    else:
        raise RuntimeError(f"Unexpected image shape: {img_bgr.shape}")

    h = gray.shape[0]
    w = gray.shape[1]
    h_1 = int((h / 4) * 4)
    w_1 = int((w / 4) * 4)
    small_gray = cv2.resize(gray, (w_1, h_1))
    reshaped_small_gray_img = small_gray.reshape(
        (1, 1, small_gray.shape[0], small_gray.shape[1]))
    return reshaped_small_gray_img


import numpy as np


def cosine_distance(encoding_to_compare: np.ndarray, encodings: np.ndarray,
                    is_normalized: bool = False):
    """Return the cosine distance of 'encoding_to_compare' to every encoding
    in 'encodings'"""
    if not is_normalized:
        # normalize arrays to unit length vectors (length 1)
        e_normed = np.linalg.norm(encodings, axis=1, keepdims=True)
        encodings = np.asarray(encodings) / e_normed

        etc_normed = np.linalg.norm(encoding_to_compare, axis=0, keepdims=True)
        encoding_to_compare = np.asarray(encoding_to_compare) / etc_normed
    return 1. - np.dot(encodings, encoding_to_compare.T)


def euclidian_distance(encoding_to_compare: np.ndarray, encodings: np.ndarray):
    """
    Given a list of face encodings, compare them to a known face encoding and
    get a euclidean distance for each comparison face. The distance tells you
    how similar the faces are.

    :param encoding_to_compare: A face encoding to compare against
    :param encodings: List of face encodings to compare
    :return: A numpy ndarray with the distance for each face in the same order
             as the 'faces' array
    """
    if len(encodings) == 0:
        return np.empty(0)
    assert encodings.shape[-1] == encoding_to_compare.shape[-1], \
        "Encoding shapes are incorrect!"
    return np.linalg.norm(encodings - encoding_to_compare, axis=1)



import logging
import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Deque
from collections import deque

import numpy as np

from vcap import Resize, BaseBackend, DetectionNode

_SUPPORTED_METRICS = "SUPPORTED_METRICS"
_RANGE_FOR_ASYNC_INFER_REQUESTS = "RANGE_FOR_ASYNC_INFER_REQUESTS"
OV_INPUT_TYPE = Dict[str, np.ndarray]


class BaseOpenVINOBackend(BaseBackend):
    def __init__(self, model_xml: bytes,
                 weights_bin: bytes,
                 device_name: str,
                 ie_core=None):
        """
        :param model_xml: The XML data defining the OpenVINO model architecture
        :param weights_bin: The .bin file data defining the model's weights
        :param ie_core: :
          None (default): The backend will initialize its own IECore to load
          the network with.
          IECore: An initialized openvino.inference_engine.IECore, with any
          settings already applied. This can be used to apply CPU extensions
          or load different plugins to the IECore giving it to the backend.
        """
        super().__init__()
        # Convert from the vcap device naming format to openvino format
        device_name = "CPU" if device_name[:4] == "CPU:" else device_name

        from openvino.inference_engine import \
            IECore, ExecutableNetwork, IENetwork, StatusCode

        self.ie = ie_core or IECore()

        # Find the optimal number of InferRequests for this device
        supported_metrics = self.ie.get_metric(
            device_name, _SUPPORTED_METRICS)
        if _RANGE_FOR_ASYNC_INFER_REQUESTS in supported_metrics:
            _, n_requests, _ = self.ie.get_metric(
                device_name, _RANGE_FOR_ASYNC_INFER_REQUESTS)
        else:
            # Use the devices default
            n_requests = 0

        self.net: IENetwork = self.ie.read_network(
            model=model_xml,
            weights=weights_bin,
            init_from_buffer=True)

        self.exec_net: ExecutableNetwork = self.ie.load_network(
            network=self.net,
            device_name=device_name,
            num_requests=n_requests)

        # Pull out a couple useful constants
        self.InferRequestStatusCode = StatusCode
        self.input_blob_names: List[str] = list(self.net.inputs.keys())
        self.output_blob_names: List[str] = list(self.net.outputs.keys())

    def prepare_inputs(self, frame: np.ndarray, frame_input_name: str = None) \
            -> Tuple[OV_INPUT_TYPE, Resize]:
        """A helper method to create an OpenVINO input like {input_name: array}

        This method takes a frame, resizes it to fit the network inputs, then
        returns two things: The input, and the Resize information. The
        Resize information contains all of the operations that were done on
        the frame, allowing users to then map the detections from a resized
        frame to the coordinate space of the original frame.

        :param frame: The image. BGR ordered.
        :param frame_input_name: Set this value to force a certain node to be
            used as the frame input. Useful if you still want to use the
            default implementation from a subclass with network with multiple
            inputs
        :returns: ({input_name: resized_frame}, Resize)
        """

        if not frame_input_name and len(self.net.inputs) > 1:
            raise ValueError("More than one input was expected for model, but "
                             "default prepare_inputs implementation was used.")

        input_blob_name = frame_input_name or self.input_blob_names[0]
        input_blob = self.net.inputs[input_blob_name]

        _, _, h, w = input_blob.shape
        resize = Resize(frame).resize(w, h, Resize.ResizeType.EXACT)

        # Change data layout from HWC to CHW
        in_frame = np.transpose(resize.frame.copy(), (2, 0, 1))

        return {input_blob_name: in_frame}, resize

    def parse_detection_results(
            self, results: np.ndarray,
            resize: Resize,
            label_map: Dict[int, str],
            min_confidence: float = 0.0) -> List[DetectionNode]:
        """A helper method to take results from a detection-type network.
        :param results: The inference results from the network
        :param resize: A Resize object that was used to resize the image to
        fit into the network originally.
        :param label_map: A dictionary mapping integers to class_names.
        :param min_confidence: Filter out detections that have a confidence
        less than this number.
        :returns: A list of DetectionNodes, in this case representing bounding
        boxes.
        """
        output_blob_name = self.output_blob_names[0]
        inference_results = results[output_blob_name]

        _, _, h, w = self.net.inputs[self.input_blob_names[0]].shape

        nodes: List[DetectionNode] = []
        for result in inference_results[0][0]:
            # If the first index == 0, that's the end of real predictions
            # The network always outputs an array of length 200 even if it does
            # not have that many predictions
            if result[0] != 0:
                break

            confidence = float(result[2])
            if confidence <= min_confidence:
                continue

            x_min, y_min, x_max, y_max = result[3:7]
            # x and y in res are in terms of percent of image width/height
            x_min, x_max = x_min * w, x_max * w
            y_min, y_max = y_min * h, y_max * h
            coords = [[x_min, y_min], [x_max, y_min],
                      [x_max, y_max], [x_min, y_max]]

            class_id = round(result[1])
            res = DetectionNode(
                name=label_map[class_id],
                coords=coords,
                extra_data={"detection_confidence": confidence})
            nodes.append(res)

        # Convert the coordinate space of the detections from the
        # resized frame to the
        resize.scale_and_offset_detection_nodes(nodes)
        return nodes

    def batch_predict(self, inputs: List[OV_INPUT_TYPE]) \
            -> List[object]:
        """Use the network for inference.

        This function will receive a list of inputs and process them as
        efficiently as possible, optimizing for throughput.
        :param inputs: A list of openvino style inputs {input_name: ndarray}
        :returns: A generator of the networks outputs, yielding them in the
        same order as the inputs
        """

        inputs: Deque[Tuple[int, OV_INPUT_TYPE]] = deque(enumerate(inputs))
        """A queue containing tuples of (frame_id, input) for inference"""
        requests_in_progress: Dict[int, int] = {}
        """A dictionary of {request_id: frame_id} for ongoing requests"""
        unsent_results: Dict[int: Dict] = {}
        """A dictionary of {frame_id: output}, the results not yet yielded"""
        next_frame_id: int = 0
        """The next frame_id we are awaiting results to send. This guarantees
        that results are sent in the same order as the inputs."""

        requests = list(enumerate(self.exec_net.requests))

        # This loop will end when all inputs have been processed and outputs
        # have been yielded
        while (len(inputs)
               + len(unsent_results)
               + len(requests_in_progress)):

            # Block until at least one request slot is free
            self.exec_net.wait(num_requests=1)

            for rid, request in requests:
                status = request.wait(0)
                if status == self.InferRequestStatusCode.INFER_NOT_STARTED:
                    # Put another request in the queue if there are frames
                    # not yet sent for processing.
                    if len(inputs):
                        frame_id, input_dict = inputs.popleft()
                        request.async_infer(input_dict)
                        requests_in_progress[rid] = frame_id

                if status != self.InferRequestStatusCode.OK:
                    # This InferRequest is currently working on a job.
                    continue

                # This request just finished an inference.
                if rid in requests_in_progress:
                    frame_id = requests_in_progress.pop(rid)
                    unsent_results[frame_id] = request.outputs

                # Put another request in the queue, if there are frames
                if len(inputs):
                    frame_id, input_dict = inputs.popleft()
                    request.async_infer(input_dict)
                    requests_in_progress[rid] = frame_id

            if next_frame_id in unsent_results:
                yield unsent_results.pop(next_frame_id)
                next_frame_id += 1

    def close(self):
        """Does nothing"""
        pass



#!/usr/bin/env python3
import os

from setuptools import setup, find_namespace_packages

test_packages = ["pytest", "mock"]

PRE_RELEASE_SUFFIX = os.environ.get("PRE_RELEASE_SUFFIX", "")

setup(
    name='vcap-utils',
    description="Utilities for creating OpenVisionCapsules easily in Python",
    author="Dilili Labs",
    packages=find_namespace_packages(include=["vcap_utils*"]),
    version="0.2.0" + PRE_RELEASE_SUFFIX,

    install_requires=[
        "vcap",
    ],

    extras_require={
        "tests": test_packages,
    },
    tests_require=test_packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import abc

import numpy as np

from vcap.backend import BaseBackend


class BaseEncoderBackend(BaseBackend):
    # TODO: Remove the concept of Backends needing a 'distances' function
    @abc.abstractmethod
    def distances(self, encoding_to_compare: np.ndarray,
                  encodings: np.ndarray) -> np.ndarray:
        """
        :param encoding_to_compare: An object encoding to compare
        :param encodings: List of object encodings to compare against
        :return: A numpy ndarray with the distance for each face in the same
        order as the 'encodings_to_compare' array
        """
        pass
